# -*- coding:utf-8 -*- 
#
# Copyright(C) 2014 Laurent Faipot (laurent.faipot@free.fr). All rights reserved.
#

import sys
import time
import os
from PyQt4 import QtGui, QtCore
from functools import partial

from boardUi import Ui_MainWindow
from assemblyUi import Ui_AssemblyWindow
from board import Board
from memory import Memory
from processor import Processor, Register, Indicator
from config import Config
from error import *

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

class BoardSimu(QtGui.QMainWindow, Ui_MainWindow):
    
    __memoryBytesPerLine = 8
    
    def displayCritical(self, exception):
        QtGui.QMessageBox.critical(
                          self,
                          'Critical',
                          str(exception)
                          )
        
    def displayAbout(self):
        QtGui.QMessageBox.about(self, "Board Simulator", 
                                self.tr(u"Version 0.2\n" +
                                "\n" +
                                "Professeur, beta-testeur, support education\n" +
                                "Sylvaine Chambre (Sylvaine.Chambre@free.fr)\n" +
                                "\n" +
                                "Concepteur, developpeur\n" +
                                "Laurent Faipot (Laurent.Faipot@free.fr)")
                                )
    
    def __init__(self, parent=None):
        super(BoardSimu, self).__init__(parent)
        self.setupUi(self)
        self.defineActions()
        self.instructionTable.setColumnWidth(0,70)
        self.instructionTable.setColumnWidth(1,70)
        self.instructionTable.setColumnWidth(2,70)
        
        Error.whenHappen(self.displayCritical)
        
    def closeEvent(self, event):
        self.quit()
        #event.accept() # let the window close
        
    def defineActions(self):
        # look for available architecture
        archList = []
        for filename in os.listdir(Config.getArchDir()):
            print "filename: " + filename
            # only file with arch_xxx.py must be considered
            print filename[-3:]
            if filename[0:5] == "arch_" and filename[-3:] == ".py":
                archList.append(filename[5:-3])
                
        # look for last used
        lastArch = config.getOption("Architecture", "last_used")
        if lastArch == "":
            self.arch = archList[0]
        else:
            self.arch = lastArch
        
        self.arch = "edu1"
        #self.actionArch = []
        #count = 0
        #for archName in archList:
        #    action = QtGui.QAction(self)
        #    print "Archname " + archName
        #    action.setObjectName(_fromUtf8("action"+ archName))
        #    f = partial(self.openArch, archName)
        #    action.triggered.connect(f)
        #    action.setText(QtGui.QApplication.translate("MainWindow", archName, None, QtGui.QApplication.UnicodeUTF8))
        #    self.menuArchitecture.addAction(action)
        #    self.actionArch.append(action)
        #    count += 1
        
        self.actionQuit.triggered.connect(self.quit)
        self.actionOpen.triggered.connect(self.openFile)
        self.actionAbout.triggered.connect(self.displayAbout)
        QtCore.QObject.connect(self.runButton, QtCore.SIGNAL('clicked()'), self.runProgram)
        QtCore.QObject.connect(self.stepButton, QtCore.SIGNAL('clicked()'), self.stepProgram)
        QtCore.QObject.connect(self.stopButton, QtCore.SIGNAL('clicked()'), self.stopProgram)
        
    def openFile(self):
        lastOpenedFile = config.getLastOpenedFile()
        if (lastOpenedFile == ""):
            lastOpenedFile = QtCore.QDir.homePath()
        fileName = QtGui.QFileDialog.getOpenFileName(
                        self,
                        "Open program file",
                        lastOpenedFile,
                        "Assembly files (*.ass *.txt);;All files (*.*)"
                        )
            
        if fileName:
            config.setRecentFiles(fileName)
            self.processFile(fileName)
            
    def openArch(self, arch):
        self.arch = arch
        
    def processFile(self, fileName):
        self.memoryTable.clear()
        self.instructionTable.clear()
        self.registerTable.clear()
        self.indicatorTable.clear()
        
        self.board = Board(self, self.arch)

        Register.whenUpdated(self.updateRegister)
        Indicator.whenUpdated(self.updateIndicator)
        Memory.whenUpdated(self.updateMemory)
        Processor.whenUpdated(self.displayInstruction)

        self.displayMemory(self.board.memory, self.board.chip.wordSize)
        self.displayRegister(self.board)
        self.displayIndicator(self.board)
        
        #self.board.program = self.board.arch.Program()
        #if (self.board.program.load(fileName, self.board)):
        if (self.board.loadProgram(fileName)):
            self.runButton.setEnabled(True)
            self.stepButton.setEnabled(True)
            self.stopButton.setEnabled(True)
            self.board.chip.PC.set(self.board.program.getCodeBase())
               
        self.initProgram(self.board.program)
    
        
    def initProgram(self, program):
        
        self.previousRow = -1
        self.stop = False
        self.end = False

        self.displayProgram(program)

        
    def runProgram(self):
        self.stop = False
        self.runButton.setEnabled(False)
        self.stepButton.setEnabled(False)
        self.stopButton.setEnabled(True)
        while (not self.stop and not self.end):
            self.executeNext(self.board)
            app.processEvents()
            # on virtualized Linux, show 1 change out of 2 if only 1 app.processEvents
            app.processEvents()
            if (not self.end):
                time.sleep(1)
        self.stopProgram()
        
                    
    def stepProgram(self):
        self.runButton.setEnabled(True)
        self.stepButton.setEnabled(True)
        self.stopButton.setEnabled(False)
 
        self.executeNext(self.board)
            
    def stopProgram(self):
        self.stop = True
        self.stopButton.setEnabled(False)
        if not self.end:
            self.runButton.setEnabled(True)
            self.stepButton.setEnabled(True)
        
    def executeNext(self, board):
        self.cleanMemoryMarker(self.board.memory)
        self.cleanRegisterMarker()
        self.cleanIndicatorMarker()
        try:
            self.end = self.board.chip.executeNext(self.board)
            return
        except Error, e:
            self.end = True
            self.stop = False
            self.runButton.setEnabled(False)
            self.stepButton.setEnabled(False)
            self.stopButton.setEnabled(False)
            e.display()
        finally:
            if (self.end):
                self.runButton.setEnabled(False)
                self.stepButton.setEnabled(False)
                self.stopButton.setEnabled(False)
        
    def start(self):
        self.show()
        sys.exit(self.app.exec_())
        
    def quit(self):
        print "QUIT requested"
        QtGui.qApp.quit()
    
    @staticmethod
    def hexa(value, length):
        nbits = length * 8
        s = hex((value + (1 << nbits)) % (1 << nbits))
        s = s[2:]
        return s.rjust(length*2, '0').upper()
    
    def displayMemory(self, memory, wordSize):
        size = memory.getSize()
        self.nbMemoryRow = size / BoardSimu.__memoryBytesPerLine
        self.nbMemoryCol = (BoardSimu.__memoryBytesPerLine / wordSize)
        self.memoryTable.setRowCount(self.nbMemoryRow)
        self.memoryTable.setColumnCount(self.nbMemoryCol +1)
        
        self.memoryTable.setColumnWidth(0,60)
        for col in range(self.nbMemoryCol):
            self.memoryTable.setColumnWidth(col + 1, 30 * wordSize)
            
        for row in range(self.nbMemoryRow):
            address = '{:04d}'.format(row * BoardSimu.__memoryBytesPerLine);
            item = QtGui.QTableWidgetItem(address)
            self.memoryTable.setItem(row, 0, item)
            col = 0
            while (col < self.nbMemoryCol):
                #item = QtGui.QTableWidgetItem(address)
                #self.memoryTable.setItem(row, col+1, item)
                mem = memory.get(row * BoardSimu.__memoryBytesPerLine + col, wordSize)
                value = BoardSimu.hexa(mem, wordSize)
                item = QtGui.QTableWidgetItem(value)
                self.memoryTable.setItem(row, col+1, item)
                col += 1
    
    def updateMemory(self, memory, address, length, wordSize = 1):        
        size = memory.getSize()
        i = address
        while (i < size) & (i < address + length):
            row = i / BoardSimu.__memoryBytesPerLine
            col = i - (row * BoardSimu.__memoryBytesPerLine)
            mem = memory.get(row * BoardSimu.__memoryBytesPerLine + col, wordSize)
            value = BoardSimu.hexa(mem, wordSize)
            self.memoryTable.item(row, col+1).setText(value)
            self.memoryTable.item(row, col+1).setBackground(QtGui.QColor(0,160,0))
            i += wordSize
    
    def cleanMemoryMarker(self, memory):
        for row in range(self.nbMemoryRow):
            for col in range(self.nbMemoryCol):
                self.memoryTable.item(row, col+1).setBackground(QtGui.QColor(255,255,255))
            
    def displayProgram(self, program):
        row = 0
        instructionList = program.getInstructionList()
        print "instructionList = " + str(instructionList)
        self.instructionTable.setRowCount(len(instructionList))
        for inst in instructionList:
            for col in range(3):
                print "inst " + str(inst)
                value = inst[col]
                if (col == 0):
                    value = str(inst[col])
                item = QtGui.QTableWidgetItem(value)
                self.instructionTable.setItem(row, col, item)
            row +=1
            
    def displayInstruction(self, pc):
        # look for instruction based on its address
        for row in range(self.instructionTable.rowCount()):
            address = int(self.instructionTable.item(row, 0).text())
            if (address >= pc):
                break
        #row = (pc - self.board.program.getCodeBase()) / (self.board.chip.instructionSize + self.board.chip.operandSize)
        #print "PC " + str(pc) + "  codeBase: " + str(self.board.program.getCodeBase()) +  " instsize: " + str(self.board.chip.instructionSize)
        self.instructionTable.selectRow(row)
        self.previousRow = row
        
    def displayRegister(self, board):
        registerList = board.chip.getRegisterList()
        self.registerTable.setRowCount(len(registerList))
        self.registerTable.setColumnCount(3)
        self.registerTable.setColumnWidth(0, 40)
        self.registerTable.setColumnWidth(1, 60)
        self.registerTable.setColumnWidth(2, 60)
        count = 0
        for reg in registerList:
            item = QtGui.QTableWidgetItem(reg.getName())
            self.registerTable.setItem(count, 0, item)
            item = QtGui.QTableWidgetItem(str(reg.get()))
            self.registerTable.setItem(count, 1, item)
            item = QtGui.QTableWidgetItem(str(reg.get()))
            self.registerTable.setItem(count, 2, item)
            count += 1
            
    def updateRegister(self, reg):
        value = reg.get()
        wordSize    = self.board.chip.wordSize
        self.registerTable.item(reg.getRank(), 1).setText(str(value))
        self.registerTable.item(reg.getRank(), 1).setBackground(QtGui.QColor(0,160,0))
        self.registerTable.item(reg.getRank(), 2).setText(BoardSimu.hexa(value, wordSize))
        self.registerTable.item(reg.getRank(), 2).setBackground(QtGui.QColor(0,160,0))
        return
    
    def displayIndicator(self, board):
        indicatorList = board.chip.getIndicatorList()
        self.indicatorTable.setRowCount(len(indicatorList))
        self.indicatorTable.setColumnCount(2)
        self.indicatorTable.setColumnWidth(0, 40)
        self.indicatorTable.setColumnWidth(1, 20)
        count = 0
        for indicator in indicatorList:
            item = QtGui.QTableWidgetItem(indicator.getName())
            self.indicatorTable.setItem(count, 0, item)
            item = QtGui.QTableWidgetItem(str(indicator.get()))
            self.indicatorTable.setItem(count, 1, item)
            count += 1
 
    def updateIndicator(self, indicator):
        value = indicator.get()
        self.indicatorTable.item(indicator.getRank(), 1).setText(str(value))
        self.indicatorTable.item(indicator.getRank(), 1).setBackground(QtGui.QColor(0,160,0))
   
    
    def cleanRegisterMarker(self):
        for row in range(self.registerTable.rowCount()):
            for col in range(self.registerTable.columnCount()):
                self.registerTable.item(row, col).setBackground(QtGui.QColor(255,255,255))
                
    def cleanIndicatorMarker(self):
        for row in range(self.indicatorTable.rowCount()):
            for col in range(self.indicatorTable.columnCount()):
                self.indicatorTable.item(row, col).setBackground(QtGui.QColor(255,255,255))
                
                
class AssemblySimu(QtGui.QMainWindow, Ui_AssemblyWindow):
    
    def displayAbout(self):
        QtGui.QMessageBox.about(self, "Board Simulator", "Version 0.1 by Lolo")

    
    def __init__(self, parent=None):
        super(AssemblySimu, self).__init__(parent)
        self.setupUi(self)
        self.connectActions()
        
    def connectActions(self):
        #self.actionQuit.triggered.connect(QtGui.qApp.quit)
        self.actionOpen.triggered.connect(self.openFile)
        #self.actionAbout.triggered.connect(self.displayAbout)
        QtCore.QObject.connect(self.loadButton, QtCore.SIGNAL('clicked()'), self.loadProgram)
        
    def openFile(self):
        lastOpenedFile = config.getLastOpenedFile()
        print "PATH_0 " + lastOpenedFile
        if (lastOpenedFile == ""):
            lastOpenedFile = Config.currentDir()
            print "HOME PATH: " + lastOpenedFile
            fileName = QtGui.QFileDialog.getOpenFileName(
                        self,
                        "Open program file",
                        lastOpenedFile,
                        "Assembly files (*.ass *.txt);;All files (*.*)"
                        )
            
        if fileName:
            print "filename: " + fileName
            config.setRecentFiles(fileName)
            self.processFile(fileName)
    
    def processFile(self, fileName):
        self.SourceText.clear()
        f = open(fileName, "r");
        for line in f:
            self.SourceText.append(line.rstrip("\n"))
            
    def saveFile(self):
        pass
    
    def loadProgram(self):
        pass

if __name__ == '__main__': 
    app = QtGui.QApplication(sys.argv)
    config = Config("boardsimu.conf")
    boardSimu = BoardSimu()
    boardSimu.show()
    sys.exit(app.exec_())
