# -*- coding:utf-8 -*- 
#
# Copyright(C) 2014 Laurent Faipot (laurent.faipot@free.fr). All rights reserved.
#
from processor import *

class Architecture:
    
    memorySize      = 512

class Chip(Processor):
    
    wordSize        = 1 # byte
    regSize         = 1 # byte
    instructionSize = 1 # byte
    operandSize     = 1
    
    def lda(self, board, operand):
        # determine operand type
        self.regA.set(operand)
        
    def tax(self, board, operand):
        self.regX.set(self.regA.get())
        
    def inx(self, board, operand):
        self.regX.set(self.regX.get() + 1)
        
    def adc(self, board, operand):
        self.regA.set(self.regA.get() + self.carry.get() + operand)
        
    def brk(self,board, operand):
        self.endProgram = True
        
    operandSet = [
        [Processor.OpNumeric, "\#\$([a-fA-F0-9]*)", Processor.OpFormatHexa, -1],
        [Processor.OpNumeric, "\#([0-9]*)", Processor.OpFormatDecimal, -1],
      ]
        
    instructionSet= {
        'LDA':  [0xa9, lda, [Processor.OpNumeric]],
        'TXA':  [0xaa, tax, [Processor.OpVoid]],
        'INX':  [0xe8, inx, [Processor.OpVoid]],
        'ADC':  [0x69, adc, [Processor.OpNumeric]],
        'BRK':  [0x00, brk, [Processor.OpVoid]]
        }
    
    def __init__(self):
        #super(Chip, self).__init__("test")
        Processor.__init__(self, "6502", Chip.instructionSize, Chip.operandSize)
        self.PC    = self.addRegister('PC', Register.PC, Register.UnsignedInteger, Chip.regSize)
        self.SP    = self.addRegister('SP', Register.SP, Register.UnsignedInteger, Chip.regSize)
        self.regA  = self.addRegister('A', Register.AccIndic, Register.SignedInteger, Chip.regSize)
        self.regX  = self.addRegister('X', Register.Index, Register.SignedInteger, Chip.regSize)
        self.regY  = self.addRegister('Y', Register.Index, Register.SignedInteger, Chip.regSize)
        
        self.carry      = self.addIndicator('C', Indicator.Carry)
        self.zero       = self.addIndicator('Z', Indicator.Zero)
        self.overflow   = self.addIndicator('O', Indicator.Overflow)
        self.negative   = self.addIndicator('N', Indicator.Negative)
        
        self.defineOperand(self.operandSet)
        self.defineInstruction(self.instructionSet)
        
class Program(Assembly):

    def __init__(self):
        Assembly.__init__(self)

    def load(self, fileName, board):
        memory = board.memory
        "load code into memory"
        f = open(fileName, "r");
        dataSection = False
        lineNo = 0
        
        error = False
        errorList = []
        
        for line in f:
            
            lineNo += 1
            
            # .DATA
            #   signale le debut des donnes suivi de ligne
            # <adresse> valeur
            #
            # .CODE <adresse>
            #    signale le debut du code avec adresse de chargement
            # chaque instruction est compose de 2 champs
            # le mnemonique d'instruction et l'operande (sauf END)
            #
            
            # remove comments
            line = line.split(';', 1)[0]
            # split using spaces/tabs as separators
            instruction = line.split()
            
            # check if the line is empty, 
            if len(instruction) == 0:
                # probably an empty line
                continue


            mnemonic = instruction[0].upper()
            operand = None

            # is there an operand
            if (len(instruction) > 1):
                operand   = instruction[1]
                print "Inst: " + mnemonic + " OPERAND FOUND: " + operand
            else:
                print "Inst: " + mnemonic + "-> NO OPERAND"
            
            if (mnemonic == '.DATA'):
                dataSection = True
                continue
                
            if (mnemonic == '.CODE'):
                dataSection = False
                # we store starting address of code
                if (operand):
                    self.setCodeBase(int(operand))
                else:
                    error = True
                    errorList.append("Line " + str(lineNo) + ": .CODE requires starting address")
                continue

            if (dataSection):
                # format: <address> <value>
                memory.set(int(mnemonic), 1, int(operand))
                continue
            
            address = self.getLoadAddress()
            errorMsg = self.loadAssembly(board, mnemonic, operand)
            if (len(errorMsg) == 0):
                if operand == None:
                    operand = 0
                self.storeInstruction(address, mnemonic, operand)
                
            else:
                errorList.append("Line " + str(lineNo) + ": " + errorMsg)
            
        # add a .END if there is none at the end of the source file
        if (mnemonic != "BRK"):
            errorMsg = self.loadAssembly(board, "BRK", 0)
            self.storeInstruction(self.loadAddress, "BRK", 0)
            
        if (error):
            raise ProgramError(Error.error, errorList)
                    
